;;; rt.el --- Common Lisp RT port to Emacs Lisp

;; Copyright (C) 2010 Didier Verna
;; Copyright (C) 1990 Massachusetts Institute of Technology, Cambridge MA.

;; Author:        Didier Verna <didier@xemacs.org>
;; Maintainer:    Didier Verna <didier@xemacs.org>
;; Created:       Tue Dec  8 09:38:04 2009
;; Last Revision: Mon Feb 15 17:07:09 2010
;; Keywords:      lisp


;; Permission to use, copy, modify, and distribute this software and its
;; documentation for any purpose and without fee is hereby granted, provided
;; that this copyright and permission notice appear in all copies and
;; supporting documentation, and that the name of M.I.T. not be used in
;; advertising or publicity pertaining to distribution of the software without
;; specific, written prior permission. M.I.T. makes no representations about
;; the suitability of this software for any purpose. It is provided "as is"
;; without express or implied warranty.

;; M.I.T. DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL
;; IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL M.I.T.
;; BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
;; WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
;; ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR
;; IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

;; This is the December 19, 1990 version of a set of tests that use the RT
;; regression tester to test itself. See the documentation of RT for a
;; discusion of how to use this file.


;;; Commentary:

;; Contents management by FCM version 0.1.

;; This is a port of Common Lisp's rt package to Emacs Lisp.

;; API differences:
;; - Since elisp doesn't provide packages, public symbols are called rt:<name>
;;   and private ones are called rt::<name>
;; - Instead of *STANDARD-OUTPUT* we use a buffer named *RT* by default, and
;;   we prompt for a file name otherwise.
;; - The API functions are made interactive.
;; - No more optimization settings parameter.
;; - No more expanded eval feature.
;; - *COMPILE-TESTS* replaced with *BYTE-COMPILE-TESTS*.

;; Internals:
;; - No more Common Lisp compiler-specific code
;; - Use of CL for as much Common Lisp emulation as possible, but some elisp
;;   idioms remain:
;;   * *print-circle-on-failures* replaced by *print-level-on-failures*
;;   * SPECIAL-FORM-P instead of special-operator-p (although a recent version
;;     of XEmacs 21.5 would be happy with s-f-p)
;;   * no MACRO-FUNCTION, but (eq 'macro (car (symbol-function sym)))
;;   * condition-case instead of handler-bind
;;   * FORMAT is not as powerful
;;   * No multidimensionnal arrays in Emacs Lisp (only vectors), and array
;;     means something else: vector, string or bit vector.


;;; Code:

(require 'cl)


(defvar rt:*test* nil
  "Current test name")
(defvar rt:*do-tests-when-defined* nil)

(defvar rt::*entries* '(nil)
  "Test database.  Has a leading dummy cell that does not contain an entry.")
(defvar rt::*entries-tail* rt::*entries*
  "Tail of the *entries* list")
(defvar rt::*entries-table* (make-hash-table :test #'equal)
  "Map the names of entries to the cons cell in *entries* that precedes the
one whose car is the entry.")
(defvar rt::*in-test* nil
  "Used by TEST")
(defvar rt::*debug* nil
  "For debugging")
(defvar rt::*catch-errors* t
  "When true, causes errors in a test to be caught.")
(defvar rt::*print-level-on-failure* nil
  "Failure reports are printed with PRINT-LEVEL bound to this value.")
(defvar rt::*byte-compile-tests* nil
  "When true, byte-compile the tests before running them.")

(defvar rt::*expected-failures* nil
  "A list of test names that are expected to fail.")

(defvar rt::*notes* (make-hash-table :test 'equal)
  "A mapping from names of notes to note objects.")

(defstruct (rt::entry (:constructor rt::make-entry)
		      (:conc-name rt::)
		      (:copier rt::copy-entry))
  pend name props form vals)

;;; Note objects are used to attach information to tests.
;;; A typical use is to mark tests that depend on a particular
;;; part of a set of requirements, or a particular interpretation
;;; of the requirements.

(defstruct (rt::note (:constructor rt::make-note)
		     (:conc-name rt::note-)
		     (:copier rt::copy-note))
  name
  contents
  disabled ;; When true, tests with this note are considered inactive
  )

;; (defmacro vals (entry) `(cdddr ,entry))

(defmacro rt::defn (entry)
  (let ((var (gensym)))
    `(let ((,var ,entry))
      (list* (rt::name ,var) (rt::form ,var) (rt::vals ,var)))))

(defun rt::entry-notes (entry)
  (let ((notes (getf (rt::props entry) :notes)))
    (if (listp notes)
	notes
      (list notes))))

(defun rt::has-disabled-note (entry)
  (let ((notes (rt::entry-notes entry)))
    (loop for n in notes
      for note = (if (rt::note-p n)
		     n
		   (gethash n rt::*notes*))
      thereis (and note (rt::note-disabled note)))))

(defun rt:pending-tests ()
  (interactive)
  (loop for entry in (cdr rt::*entries*)
    when (and (rt::pend entry) (not (rt::has-disabled-note entry)))
    collect (rt::name entry)))

(defun rt:rem-all-tests ()
  (interactive)
  (setq rt::*entries* (list nil))
  (setq rt::*entries-tail* rt::*entries*)
  (clrhash rt::*entries-table*)
  (let ((buffer (get-buffer "*RT*")))
    (when buffer
      (erase-buffer buffer)))
  nil)

(defun* rt:rem-test (&optional (name rt:*test*))
  ;; #### FIXME: should prompt for a test name with completion
  (interactive "xTest name: ")
  (let ((pred (gethash name rt::*entries-table*)))
    (when pred
      (if (null (cddr pred))
	  (setq rt::*entries-tail* pred)
	(setf (gethash (rt::name (caddr pred)) rt::*entries-table*) pred))
      (setf (cdr pred) (cddr pred))
      (remhash name rt::*entries-table*)
      name)))

(defun rt::report-error (error? &rest args)
  (cond (rt::*debug*
	 (with-current-buffer (get-buffer-create "*RT*")
	   (insert (apply #'format args)))
	 (if error?
	     (throw 'rt::*debug* nil)))
	(error?
	 (apply #'error args))
	(t
	 (apply #'warn args)))
  nil)

(defun rt::get-entry (name)
  (let ((entry (cadr (gethash name rt::*entries-table*))))
    (when (null entry)
      (rt::report-error t "No test with name %s." name))
    entry))

(defun* rt:get-test (&optional (name rt:*test*))
  ;; #### FIXME: should prompt for a test name with completion
  (interactive "xTest name: ")
  (rt::defn (rt::get-entry name)))

(defun rt::add-entry (entry)
  (setq entry (rt::copy-entry entry))
  (let* ((pred (gethash (rt::name entry) rt::*entries-table*)))
    (cond
     (pred
      (setf (cadr pred) entry)
      (rt::report-error nil "Redefining test %s." (rt::name entry)))
     (t
      (setf (gethash (rt::name entry) rt::*entries-table*) rt::*entries-tail*)
      (setf (cdr rt::*entries-tail*) (cons entry nil))
      (setf rt::*entries-tail* (cdr rt::*entries-tail*)))))
  (when rt:*do-tests-when-defined*
    (rt::do-entry entry))
  (setq rt:*test* (rt::name entry)))

(eval-when (eval compile load) (put 'rt:deftest 'lisp-indent-function 1))
(defmacro rt:deftest (name &rest body)
  (let* ((p body)
	 (properties
	  (loop while (keywordp (first p))
	    unless (cadr p)
	    do (error "Poorly formed deftest: %s."
		      (list* 'rt:deftest name body))
	    append (list (pop p) (pop p))))
	 (form (pop p))
	 (vals p))
    `(rt::add-entry (rt::make-entry :pend t
				    :name ',name
				    :props ',properties
				    :form ',form
				    :vals ',vals))))

(defun* rt:do-test (&optional (name rt:*test*))
  ;; #### FIXME: should prompt for a test name with completion
  (interactive "xTest name: ")
  (rt::do-entry (rt::get-entry name))
  (switch-to-buffer "*RT*"))

(defun rt::my-aref (a &rest args)
  (apply #'aref a args))

(defun rt::equalp-with-case (x y)
  "Like EQUALP, but doesn't do case conversion of characters."
  (cond
   ((eq x y) t)
   ((consp x)
    (and (consp y)
	 (rt::equalp-with-case (car x) (car y))
	 (rt::equalp-with-case (cdr x) (cdr y))))
   ((typep x 'vector)
    (and (typep y 'vector)
	 (let ((x-len (length x))
	       (y-len (length y)))
	   (and (eql x-len y-len)
		(loop
		  for i from 0 below x-len
		  for e1 = (rt::my-aref x i)
		  for e2 = (rt::my-aref y i)
		  always (rt::equalp-with-case e1 e2))))))
   ((and (typep x 'array)
	 (typep y 'array)
	 (not (= (length x) (length y))))
    nil)
   ((typep x 'array)
    (and (typep y 'array)
	 (let ((size (length x)))
	   (loop for i from 0 below size
	     always (rt::equalp-with-case (rt::my-aref x i)
					  (rt::my-aref y i))))))
   (t (eql x y))))

(defun* rt::do-entry (entry &optional (buffer (get-buffer-create "*RT*")))
  (catch 'rt::*in-test*
    (setq rt:*test* (rt::name entry))
    (setf (rt::pend entry) t)
    (let* ((rt::*in-test* t)
	   ;; (*break-on-warnings* t)
	   (aborted nil)
	   r)
      ;; (declare (special *break-on-warnings*))
      (block aborted
	(setf r
	      (flet ((%do ()
		       (cond
			(rt::*byte-compile-tests*
			 (multiple-value-list
			  (funcall (byte-compile `(lambda ()
						    ,(rt::form entry))))))
			(t
			 (multiple-value-list
			  (eval (rt::form entry)))))))
		(if rt::*catch-errors*
		    (condition-case c
			(%do)
		      (error
		       (setf aborted t)
		       (setf r (list c))
		       (return-from aborted nil)))
		  (%do)))))
      (setf (rt::pend entry)
	    (or aborted
		(not (rt::equalp-with-case r (rt::vals entry)))))
      (when (rt::pend entry)
	(with-current-buffer buffer
	  (end-of-buffer)
	  (let ((print-level rt::*print-level-on-failure*))
	    (insert (format "\
Test %s failed
Form: %s
Expected value(s): %s\n"
			    rt:*test*
			    (rt::form entry)
			    (rt::vals entry)))
	    ;; #### NOTE: 3 spaces here for output alignment:
	    (insert "Actual value(s):   ")
	    (condition-case nil
		(insert (format "%s\n\n" r))
	      (error
	       (insert "#<error during printing>\n\n"))))))))
  (when (not (rt::pend entry))
    rt:*test*))

(defun rt::do-entries (buffer)
  (switch-to-buffer buffer)
  (end-of-buffer)
  (insert (format "Doing %s pending test(s) of %s test(s) total.\n\n"
		  (count t (cdr rt::*entries*) :key #'rt::pend)
		  (length (cdr rt::*entries*))))
  (dolist (entry (cdr rt::*entries*))
    (when (and (rt::pend entry)
	       (not (rt::has-disabled-note entry)))
      (let ((result (rt::do-entry entry buffer)))
	(when result
	  (insert (format "%s\n\n" result))))))
  (let ((pending (rt:pending-tests))
	(expected-table (make-hash-table :test #'equal)))
    (dolist (ex rt::*expected-failures*)
      (setf (gethash ex expected-table) t))
    (let ((new-failures
	   (loop for pend in pending
	     unless (gethash pend expected-table)
	     collect pend)))
      (if (null pending)
	  (insert "No tests failed.\n")
	(progn
	  (insert (format "%s out of %s total test(s) failed: %s\n"
			  (length pending)
			  (length (cdr rt::*entries*))
			  pending))
	  (if (null new-failures)
	      (insert "No unexpected failures.\n")
	    (when rt::*expected-failures*
	      (insert (format "%s unexpected failures: %s\n"
			      (length new-failures)
			      new-failures))))
	  (insert "\n")))
      (null pending))))

(defun rt:continue-testing ()
  (interactive)
  (if rt::*in-test*
      (throw 'rt::*in-test* nil)
    (rt::do-entries (get-buffer-create "*RT*"))))

(defun* rt:do-tests (&optional (buffer (get-buffer-create "*RT*")))
  (interactive (list (get-buffer-create
		      (if current-prefix-arg
			  (read-buffer "Buffer name: " "*RT*")
			"*RT*"))))
  (dolist (entry (cdr rt::*entries*))
    (setf (rt::pend entry) t))
  (switch-to-buffer buffer)
  (erase-buffer)
  (rt::do-entries buffer))


;;; Note handling functions and macros

(defmacro rt:defnote (name contents &optional disabled)
  `(eval-when (:load :execute)
    (let ((note (rt::make-note :name ',name
			       :contents ',contents
			       :disabled ',disabled)))
      (setf (gethash (rt::note-name note) rt::*notes*) note)
      note)))

(defun rt:disable-note (n)
  (let ((note (if (rt::note-p n) n
		(setf n (gethash n rt::*notes*)))))
    (unless note
      (error "%s is not a note or note name." n))
    (setf (rt::note-disabled note) t)
    note))

(defun rt:enable-note (n)
  (let ((note (if (rt::note-p n) n
		(setf n (gethash n rt::*notes*)))))
    (unless note
      (error "%s is not a note or note name." n))
    (setf (rt::note-disabled note) nil)
    note))

;; Emacs Lisp addition:
(defconst rt:version "1.0"
  "The current version of RT.")

(defun rt:version ()
  "Print the current version of RT."
  (interactive)
  (message "RT version %s." rt:version))


(provide 'rt)

;;; rt.el ends here
